import json
import jax
import jax.numpy as jnp
from jax.tree_util import tree_flatten

import equinox as eqx

import exciting_environments as excenvs
from dmpe.models import NeuralEulerODE


@eqx.filter_jit
def simulate_ahead(model: NeuralEulerODE, init_obs: jnp.ndarray, actions: jnp.ndarray, tau: float) -> jnp.ndarray:
    """Uses the given model to look ahead and simulate future observations.

    Args:
        model: The model to use in the simulation
        init_obs: The initial observation from which to start the simulation
        actions: The actions to apply in each step of the simulation, the length
            of the first dimension of this array determine the lenght of the
            output.
        tau: The sampling time for the model

    Returns:
        observations: The gathered observations. The shape of this is given as
            (n_actions + 1, obs_dim). That is because the first observation is
            already given through the initial observation
    """
    return model(init_obs, actions, tau)


@eqx.filter_jit
def simulate_ahead_with_env(
    env,
    init_obs: jnp.ndarray,
    init_state: jnp.ndarray,
    actions: jnp.ndarray,
) -> jnp.ndarray:
    """Uses the given environment to look ahead and simulate future observations.
    This is used to have perfect predictions

    Args:
        model: The model to use in the simulation
        init_obs: The initial observation from which to start the simulation
        init_state: The initial state from which to start the simulation
        actions: The actions to apply in each step of the simulation, the length
            of the first dimension of this array determine the lenght of the
            output.
        env_state_normalizer: Values for state normalization
        action_normalizer: Values for action normalization
        static_params: Static parameters of the environment

    Returns:
        observations: The gathered observations. The shape of this is given as
            (n_actions + 1, obs_dim). That is because the first observation is
            already given through the initial observation
    """

    action_denormalizer = jnp.array(tree_flatten(env.env_properties.action_constraints)[0], dtype=jnp.float32)

    def body_fun(carry, action):
        obs, state = carry

        state = env._ode_solver_step(state, action * action_denormalizer, env.env_properties.static_params)
        obs = env.generate_observation(state, env.env_properties.physical_constraints)
        return (obs, state), obs

    (_, last_state), observations = jax.lax.scan(body_fun, (init_obs, init_state), actions)
    observations = jnp.concatenate([init_obs[None, :], observations], axis=0)

    return observations, last_state


class ModelEnvWrapperFluidTank(eqx.Module):
    env: excenvs.CoreEnvironment

    def __call__(self, init_obs, actions, tau):

        # assumes full observability
        # assumes tau == env.tau
        # practical denormilization for fluid tank yikes
        # TODO: sort this out, error-prone
        init_state = self.env.State(
            physical_state=self.env.PhysicalState((init_obs + 1) * 3 / 2),
            PRNGKey=None,
            optional=None,
        )

        observations, _ = simulate_ahead_with_env(self.env, init_obs, init_state, actions)
        return observations


class ModelEnvWrapperPendulum(eqx.Module):
    env: excenvs.CoreEnvironment

    def __call__(self, init_obs, actions, tau):

        # assumes full observability
        # assumes tau == env.tau
        # practical denormilization for fluid tank yikes
        # TODO: sort this out, error-prone
        init_state = self.env.State(
            physical_state=self.env.PhysicalState(
                theta=init_obs[..., 0] * self.env.env_properties.physical_constraints.theta,
                omega=init_obs[..., 1] * self.env.env_properties.physical_constraints.omega,
            ),
            PRNGKey=None,
            optional=None,
        )

        observations, _ = simulate_ahead_with_env(self.env, init_obs, init_state, actions)
        return observations


class ModelEnvWrapperCartPole(eqx.Module):
    env: excenvs.CoreEnvironment

    def __call__(self, init_obs, actions, tau):

        init_state = self.env.State(
            physical_state=self.env.PhysicalState(
                init_obs[..., 0] * self.env.env_properties.physical_constraints.deflection,
                init_obs[..., 1] * self.env.env_properties.physical_constraints.velocity,
                init_obs[..., 2] * self.env.env_properties.physical_constraints.theta,
                init_obs[..., 3] * self.env.env_properties.physical_constraints.omega,
            ),
            PRNGKey=None,
            optional=None,
        )
        observations, _ = simulate_ahead_with_env(self.env, init_obs, init_state, actions)
        return observations


class ModelWrapperMassSpringDamper(eqx.Module):
    env: excenvs.CoreEnvironment

    def __call__(self, init_obs, actions, tau):
        init_state = self.env.State(
            physical_state=self.env.PhysicalState(
                init_obs[..., 0] * self.env.env_properties.physical_constraints.deflection,
                init_obs[..., 1] * self.env.env_properties.physical_constraints.velocity,
            ),
            PRNGKey=None,
            optional=None,
        )
        observations, _ = simulate_ahead_with_env(self.env, init_obs, init_state, actions)
        return observations


def save_model(filename, hyperparams, model):
    with open(filename, "wb") as f:
        hyperparam_str = json.dumps(hyperparams)
        f.write((hyperparam_str + "\n").encode())
        eqx.tree_serialise_leaves(f, model)


def load_model(filename, model_class):
    with open(filename, "rb") as f:
        hyperparams = json.loads(f.readline().decode())
        hyperparams["key"] = jnp.array(hyperparams["key"], dtype=jnp.uint32)
        model = model_class(**hyperparams)
        return eqx.tree_deserialise_leaves(f, model)



import jax
import jax.numpy as jnp


@jax.jit
def KLDLoss(p: jnp.ndarray, q: jnp.ndarray):
    """Computes the sample KLD between two inputs.

    The last dim of the input needs to be of length 1. The summation occurs along the second to
    last dimension. All dimensions before that are kept as they are. Overall the shape of the
    two inputs must be identical.
    """
    assert p.shape == q.shape, "The two inputs need to be of the same shape."
    assert p.shape[-1] == q.shape[-1] == 1, "Last dim needs to be of length 1 for PDFs"

    eps = 1e-12

    kld = (p + eps) * jnp.log((p + eps) / (q + eps))
    return jnp.squeeze(jnp.sum(kld, axis=-2))


@jax.jit
def JSDLoss(p: jnp.ndarray, q: jnp.ndarray):
    """Computes the sample JSD between two inputs.

    The last dim of the input needs to be of length 1. The summation occurs along the second to
    last dimension. All dimensions before that are kept as they are. Overall the shape of the
    two inputs must be indentical.
    """
    assert p.shape == q.shape, "The two inputs need to be of the same shape."
    assert p.shape[-1] == q.shape[-1] == 1, "Last dim needs to be of length 1 for PDFs"

    m = (p + q) / 2
    return jnp.squeeze((KLDLoss(p, m) + KLDLoss(q, m)) / 2)


def MNNS_without_penalty(data_points: jnp.ndarray, new_data_points: jnp.ndarray) -> jnp.ndarray:
    """From [Smits2024].

    Implementation inspired by https://github.com/google/jax/discussions/9813

    TODO: Not sure about this penalty. Seems difficult to use for continuous action-spaces?
    They used quantized amplitude levels in their implementation.
    """
    L = new_data_points.shape[0]
    distance_matrix = jnp.linalg.norm(data_points[:, None, :] - new_data_points[None, ...], axis=-1)
    minimal_distances = jnp.min(distance_matrix, axis=0)
    return -jnp.sum(minimal_distances) / L


def audze_eglais(data_points: jnp.ndarray, eps: float = 0.001) -> jnp.ndarray:
    """From [Smits2024]. The maximin-design penalizes points that
    are too close in the point distribution.

    TODO: There has to be a more efficient way to do this.
    """
    N = data_points.shape[0]
    distance_matrix = jnp.linalg.norm(data_points[:, None, :] - data_points[None, ...], axis=-1)
    distances = distance_matrix[jax.numpy.triu_indices(N, k=1)]

    return 2 / (N * (N - 1)) * jnp.sum(1 / (distances**2 + eps))


@jax.jit
def MC_uniform_sampling_distribution_approximation(
    data_points: jnp.ndarray, support_points: jnp.ndarray
) -> jnp.ndarray:
    """From [Smits2024]. The minimax-design tries to minimize
    the distances of the data points to the support points.

    What stops the data points to just flock to a single support point?
    This is just looking at the shortest distance.
    """
    M = support_points.shape[0]
    distance_matrix = jnp.linalg.norm(data_points[:, None, :] - support_points[None, ...], axis=-1)
    minimal_distances = jnp.min(distance_matrix, axis=0)

    return jnp.sum(minimal_distances) / M


def blockwise_mcudsa(data_points: jnp.ndarray, support_points: jnp.ndarray) -> jnp.ndarray:
    """Blockwise implementation of MCUDSA. For long trajectories, the full computation is infeasible and
    needs to be split up into smaller blocks."""

    M = support_points.shape[0]
    block_size = 1_000
    value = jnp.zeros(1)

    for m in range(0, M, block_size):
        end = min(m + block_size, M)  # next block or until the end
        value = value + (
            MC_uniform_sampling_distribution_approximation(
                data_points=data_points,
                support_points=support_points[m:end],
            )
            * (end - m)  # denormalizing mean inside loss computation
            / M
        )

    return value


@jax.jit
def kiss_space_filling_cost(
    data_points: jnp.ndarray,
    support_points: jnp.ndarray,
    variances: jnp.ndarray,
    eps: float = 1e-16,
) -> jnp.ndarray:
    """From [Kiss2024]. Slightly modified to use the mean instead of the sum in the denominator.
    The goal is to have the same metric value for identical data distributions with different number
    of data points.
    """
    difference = data_points[None, ...] - support_points[:, None, :]
    exponent = -0.5 * jnp.sum(difference**2 * 1 / variances, axis=-1)

    denominator = eps + jnp.mean(jnp.exp(exponent), axis=-1)

    return jnp.mean(1 / denominator, axis=0)


def blockwise_ksfc(
    data_points: jnp.ndarray,
    support_points: jnp.ndarray,
    variances: jnp.ndarray,
    eps: float = 1e-16,
) -> jnp.ndarray:
    """Blockwise implementation of MCUDSA. For long trajectories, the full computation is infeasible and
    needs to be split up into smaller blocks."""

    M = support_points.shape[0]
    block_size = 1_000
    value = jnp.zeros(1)

    for m in range(0, M, block_size):
        end = min(m + block_size, M)  # next block or until the end
        value = value + (
            kiss_space_filling_cost(
                data_points=data_points,
                support_points=support_points[m:end],
                variances=variances,
                eps=eps,
            )
            * (end - m)  # denormalizing mean inside loss computation
            / M
        )

    return value



import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import numpy as np

import jax.numpy as jnp

from dmpe.models.model_utils import simulate_ahead


def plot_sequence(observations, actions, tau, obs_labels, action_labels, fig=None, axs=None, dotted=False):
    """Plots a given sequence of observations and actions."""

    if fig is None or axs is None:
        fig, axs = plt.subplots(nrows=1, ncols=3, figsize=(18, 6))

    t = jnp.linspace(0, observations.shape[0] - 1, observations.shape[0]) * tau

    for observation_idx in range(observations.shape[-1]):
        axs[0].plot(
            t,
            jnp.squeeze(observations[..., observation_idx]),
            "." if dotted else "-",
            markersize=1,
            label=obs_labels[observation_idx],
        )

    axs[0].title.set_text("observations, timeseries")
    axs[0].legend()
    axs[0].set_ylabel(r"$x$")
    axs[0].set_xlabel("$t$ in seconds")

    if observations.shape[-1] == 2:
        axs[1].scatter(jnp.squeeze(observations[..., 0]), jnp.squeeze(observations[..., 1]), s=1)
        axs[1].title.set_text("observation plane")
        axs[1].set_ylabel(obs_labels[1])
        axs[1].set_xlabel(obs_labels[0])
    elif observations.shape[-1] > 2:
        axs[1].scatter(jnp.squeeze(observations[..., -2]), jnp.squeeze(observations[..., -1]), s=1)
        axs[1].title.set_text("observation plane, last two obs")
        axs[1].set_ylabel(obs_labels[-1])
        axs[1].set_xlabel(obs_labels[-2])

    if actions is not None:
        if observations.shape[0] == actions.shape[0] + 1:
            observations_ = observations[:-1]
            t = t[:-1]

        if observations.shape[-1] == 1 and actions.shape[-1] == 1:
            axs[1].scatter(jnp.squeeze(actions[..., 0]), jnp.squeeze(observations_[..., 0]), s=1)
            axs[1].title.set_text("observation $\\times$ action plane")
            axs[1].set_ylabel(obs_labels[0])
            axs[1].set_xlabel(action_labels[0])

        for action_idx in range(actions.shape[-1]):
            axs[2].plot(t, jnp.squeeze(actions[..., action_idx]), label=action_labels[action_idx])

        axs[2].title.set_text("actions, timeseries")
        axs[2].legend()
        axs[2].set_ylabel(r"$u$")
        axs[2].set_xlabel(r"$t$ in seconds")

    for ax in axs:
        ax.grid(True)

    fig.tight_layout()
    return fig, axs


def plot_model_performance(model, true_observations, actions, tau, obs_labels, action_labels):
    """Compare the performance of the model to the ground truth data."""

    fig, axs = plot_sequence(
        observations=true_observations, actions=actions, tau=tau, obs_labels=obs_labels, action_labels=action_labels
    )

    pred_observations = simulate_ahead(model, true_observations[0, :], actions, tau)

    fig, axs = plot_sequence(
        observations=pred_observations,
        actions=None,
        tau=tau,
        obs_labels=obs_labels,
        action_labels=action_labels,
        fig=fig,
        axs=axs,
        dotted=True,
    )
    return fig, axs


def append_predictions_to_sequence_plot(
    fig, axs, starting_step, pred_observations, proposed_actions, tau, obs_labels, action_labels
):
    """Appends the future predictions to the given plot."""

    t = jnp.linspace(0, pred_observations.shape[0] - 1, pred_observations.shape[0]) * tau
    t += tau * starting_step  # start where the trajectory left off

    colors = list(mcolors.CSS4_COLORS.values())[: pred_observations.shape[-1]]
    for observation_idx, color in zip(range(pred_observations.shape[-1]), colors):
        axs[0].plot(
            t,
            jnp.squeeze(pred_observations[..., observation_idx]),
            color=color,
            label="pred " + obs_labels[observation_idx],
        )

    if pred_observations.shape[-1] == 2:
        axs[1].scatter(
            jnp.squeeze(pred_observations[..., 0]),
            jnp.squeeze(pred_observations[..., 1]),
            s=1,
            color=mcolors.CSS4_COLORS["orange"],
        )
    elif pred_observations.shape[-1] > 2:
        axs[1].scatter(
            jnp.squeeze(pred_observations[..., -2]),
            jnp.squeeze(pred_observations[..., -1]),
            s=1,
            color=mcolors.CSS4_COLORS["orange"],
        )
    elif pred_observations.shape[-1] == 1 and proposed_actions.shape[-1] == 1:
        axs[1].scatter(
            jnp.squeeze(proposed_actions[..., 0]),
            jnp.squeeze(pred_observations[:-1, 0]),
            s=1,
            color=mcolors.CSS4_COLORS["orange"],
        )

    colors = list(mcolors.CSS4_COLORS.values())[: pred_observations.shape[-1]]
    for action_idx, color in zip(range(proposed_actions.shape[-1]), colors):
        axs[2].plot(
            t[:-1],
            jnp.squeeze(proposed_actions[..., action_idx]),
            color=color,
            label="pred " + action_labels[action_idx],
        )

    return fig, axs


def plot_sequence_and_prediction(
    observations, actions, tau, obs_labels, actions_labels, model, init_obs, proposed_actions
):
    """Plots the current trajectory and appends the predictions from the optimization."""

    fig, axs = plot_sequence(
        observations=observations,
        actions=actions,
        tau=tau,
        obs_labels=obs_labels,
        action_labels=actions_labels,
    )

    pred_observations = simulate_ahead(model=model, init_obs=init_obs, actions=proposed_actions, tau=tau)

    fig, axs = append_predictions_to_sequence_plot(
        fig=fig,
        axs=axs,
        starting_step=observations.shape[0],
        pred_observations=pred_observations,
        proposed_actions=proposed_actions,
        tau=tau,
        obs_labels=obs_labels,
        action_labels=actions_labels,
    )

    return fig, axs


def plot_2d_kde_as_contourf(p_est, x, observation_labels):

    fig, axs = plt.subplots(figsize=(6.75, 6))

    grid_len_per_dim = int(np.sqrt(x.shape[0]))
    x_plot = x.reshape((grid_len_per_dim, grid_len_per_dim, 2))

    cax = axs.contourf(
        x_plot[..., 0],
        x_plot[..., 1],
        p_est.reshape(x_plot.shape[:-1]),
        antialiased=False,
        levels=50,
        alpha=0.9,
        cmap=plt.cm.coolwarm,
    )
    axs.set_xlabel(observation_labels[0])
    axs.set_ylabel(observation_labels[1])

    return fig, axs, cax


def plot_2d_kde_as_surface(p_est, x, observation_labels):

    fig = plt.figure(figsize=(6, 6))
    axs = fig.add_subplot(111, projection="3d")

    grid_len_per_dim = int(np.sqrt(x.shape[0]))
    x_plot = x.reshape((grid_len_per_dim, grid_len_per_dim, 2))

    _ = axs.plot_surface(
        x_plot[..., 0],
        x_plot[..., 1],
        p_est.reshape(x_plot.shape[:-1]),
        antialiased=False,
        alpha=0.8,
        cmap=plt.cm.coolwarm,
    )
    axs.set_xlabel(observation_labels[0])
    axs.set_ylabel(observation_labels[1])

    return fig, axs


def plot_metrics_by_sequence_length(results_by_metric, lengths):
    metric_keys = results_by_metric.keys()

    fig, axs = plt.subplots(3, figsize=(16, 12))
    colors = plt.rcParams["axes.prop_cycle"]()

    for metric_idx, metric_key in enumerate(metric_keys):
        axs[metric_idx].plot(
            lengths,
            jnp.array(results_by_metric[metric_key]),
            label=metric_key,
            color=next(colors)["color"],
        )

    [ax.grid(True) for ax in axs]
    [ax.legend() for ax in axs]

    return fig


def plot_mean_and_std_by_sequence_length(all_results_by_metric, lengths, use_log=False):
    metric_keys = all_results_by_metric.keys()

    fig, axs = plt.subplots(3, figsize=(16, 12), sharex=True)
    colors = plt.rcParams["axes.prop_cycle"]()

    for metric_idx, metric_key in enumerate(metric_keys):
        mean = jnp.mean(all_results_by_metric[metric_key], axis=0)
        std = jnp.std(all_results_by_metric[metric_key], axis=0)

        c = next(colors)["color"]

        axs[metric_idx].plot(
            lengths,
            jnp.log(mean) if use_log else mean,
            label=("log " if use_log else "") + metric_key,
            color=c,
        )
        axs[metric_idx].fill_between(
            lengths,
            jnp.log(mean - std) if use_log else mean - std,
            jnp.log(mean + std) if use_log else mean + std,
            color=c,
            alpha=0.2,
        )
        # axs[metric_idx].set_ylabel(("log " if use_log else "") + metric_key)
    axs[-1].set_xlabel("timesteps")
    axs[-1].set_xlim(lengths[0] - 100, lengths[-1] + 100)
    [ax.grid(True) for ax in axs]
    [ax.legend() for ax in axs]
    plt.tight_layout()

    return fig


def plot_metrics_by_sequence_length_for_all_algos(data_per_algo, lengths, algo_names, use_log=False):
    assert len(data_per_algo) == len(algo_names), "Mismatch in number of algo results and number of algo names"

    metric_keys = data_per_algo[0].keys()

    fig, axs = plt.subplots(3, figsize=(16, 12), sharex=True)
    colors = plt.rcParams["axes.prop_cycle"]()

    for algo_name, data in zip(algo_names, data_per_algo):
        c = next(colors)["color"]

        for metric_idx, metric_key in enumerate(metric_keys):
            mean = jnp.mean(data[metric_key], axis=0)
            std = jnp.std(data[metric_key], axis=0)

            axs[metric_idx].plot(
                lengths,
                jnp.log(mean) if use_log else mean,
                label=algo_name,
                color=c,
            )
            axs[metric_idx].fill_between(
                lengths,
                jnp.log(mean - std) if use_log else mean - std,
                jnp.log(mean + std) if use_log else mean + std,
                color=c,
                alpha=0.2,
            )
            # axs[metric_idx].set_ylabel(("log " if use_log else "") + metric_key)

    for idx, metric_key in enumerate(metric_keys):
        axs[idx].title.set_text(("log " if use_log else "") + metric_key)

    axs[-1].set_xlabel("timesteps")
    axs[-1].set_xlim(lengths[0] - 100, lengths[-1] + 100)
    [ax.grid(True) for ax in axs]
    [ax.legend() for ax in axs]
    plt.tight_layout()

    return fig



from typing import Callable
from functools import partial

import jax
import jax.numpy as jnp
import optax
import equinox as eqx

from exciting_environments.core_env import CoreEnvironment

from dmpe.models.model_utils import simulate_ahead
from dmpe.utils.density_estimation import (
    DensityEstimate,
    update_density_estimate_single_observation,
    update_density_estimate_multiple_observations,
)
from dmpe.utils.metrics import JSDLoss


def soft_penalty(a, a_max=1, penalty_order=2):
    """Computes penalty for the given input. Assumes symmetry in all dimensions."""
    penalties = jax.nn.relu(jnp.abs(a) - a_max)

    penalties = penalties**penalty_order

    penalty = jnp.sum(penalties, axis=(-2, -1))
    return jnp.squeeze(penalty)


@eqx.filter_jit
def loss_function(
    model,
    init_obs: jnp.ndarray,
    actions: jnp.ndarray,
    density_estimate: DensityEstimate,
    tau: float,
    target_distribution: jnp.ndarray,
    rho_obs: float,
    rho_act: float,
    penalty_order: int,
) -> jnp.ndarray:
    """Predicts a trajectory based on the given actions and the model and computes the
    corresponding loss value.

    Args:
        model: The model to use for the prediction
        init_obs: The initial observation from which to start the simulation
        actions: The actions to apply in each step of the simulation, the length
            of the first dimension of this array determine the lenght of the
            output.
        density_estimate: The current estimate of the data density
        tau: The sampling time for the model
        target_distribution: The goal distribution of the data. The JSD loss is computed
            w.r.t. this distribution
    """
    observations = simulate_ahead(model=model, init_obs=init_obs, actions=actions, tau=tau)

    predicted_density_estimate = update_density_estimate_multiple_observations(
        density_estimate, jnp.concatenate([observations[0:-1, :], actions], axis=-1)
    )
    loss = JSDLoss(
        p=predicted_density_estimate.p / jnp.sum(predicted_density_estimate.p),
        q=target_distribution / jnp.sum(target_distribution),
    )
    penalty_terms = (
        rho_obs * soft_penalty(a=observations, a_max=1, penalty_order=penalty_order)
        + rho_act * soft_penalty(a=actions, a_max=1, penalty_order=penalty_order)
        # + 5e-2 * jnp.sum(jnp.diff(actions, axis=0) ** 2)
    )

    return loss + penalty_terms


@eqx.filter_jit
def optimize_actions(
    loss_function,
    grad_loss_function,
    proposed_actions,
    model,
    optimizer,
    init_obs,
    density_estimate,
    n_opt_steps,
    tau,
    target_distribution,
    rho_obs,
    rho_act,
    penalty_order,
):
    """Uses the model to compute the effect of actions onto the observation trajectory to
    optimize the actions w.r.t. the given (gradient of the) loss function.
    """
    opt_state = optimizer.init(proposed_actions)

    def body_fun(i, carry):
        proposed_actions, opt_state = carry
        value, grad = grad_loss_function(
            model,
            init_obs,
            proposed_actions,
            density_estimate,
            tau,
            target_distribution,
            rho_obs,
            rho_act,
            penalty_order,
        )
        updates, opt_state = optimizer.update(grad, opt_state, proposed_actions)
        proposed_actions = optax.apply_updates(proposed_actions, updates)

        # proposed_actions = proposed_actions - lr * grad

        return (proposed_actions, opt_state)

    proposed_actions, _ = jax.lax.fori_loop(0, n_opt_steps, body_fun, (proposed_actions, opt_state))

    loss = loss_function(
        model,
        init_obs,
        proposed_actions,
        density_estimate,
        tau,
        target_distribution,
        rho_obs,
        rho_act,
        penalty_order,
    )

    return proposed_actions, loss


def optimize_actions_multistart(
    loss_function,
    grad_loss_function,
    all_proposed_actions,
    model,
    optimizer,
    init_obs,
    density_estimate,
    n_opt_steps,
    tau,
    target_distribution,
    rho_obs,
    rho_act,
    penalty_order,
):
    """Uses the model to compute the effect of actions onto the observation trajectory to
    optimize the actions w.r.t. the given (gradient of the) loss function. This is done for
    multiple sets of proposed actions and the best set is returned.

    """
    assert all_proposed_actions.ndim == 3, "proposed_actions must have shape (n_starts, n_opt_steps, action_dim)"

    all_optimized_actions, all_losses = jax.vmap(
        optimize_actions,
        in_axes=(None, None, 0, None, None, None, None, None, None, None, None, None, None),
    )(
        loss_function,
        grad_loss_function,
        all_proposed_actions,
        model,
        optimizer,
        init_obs,
        density_estimate,
        n_opt_steps,
        tau,
        target_distribution,
        rho_obs,
        rho_act,
        penalty_order,
    )

    best_idx = jnp.argmin(all_losses)
    return all_optimized_actions[best_idx], all_losses[best_idx]


class Exciter(eqx.Module):
    """A class that carries the necessary tools for excitation input computations.

    Args:
        grad_loss_function: The gradient of the loss function w.r.t. the actions as
            a callable function
        excitiation_optimizer: The optimizer for the excitation input computation
        n_opt_steps: Number of SGD steps per iteration
        tau: The time step length of the simulation
        target_distribution: The targeted distribution for the data density
        rho_obs: Weighting factor for observation soft constraints
        rho_act: Weighting factor for action soft constraints
    """

    loss_function: Callable
    grad_loss_function: Callable
    excitation_optimizer: optax._src.base.GradientTransformationExtraArgs
    tau: float
    n_opt_steps: int
    target_distribution: jnp.ndarray
    rho_obs: float
    rho_act: float
    penalty_order: int
    clip_action: bool
    n_starts: int
    reuse_proposed_actions: bool

    @eqx.filter_jit
    def choose_action(
        self,
        obs: jnp.ndarray,
        model,
        density_estimate: DensityEstimate,
        proposed_actions: jnp.ndarray,
        expl_key: jax.random.PRNGKey,
    ) -> tuple[jnp.ndarray, jnp.ndarray, DensityEstimate]:
        """Chooses the next action to take, updates the density estimate and
        proposes future actions.

        Args:
            obs: The current observations from which to start
            model: The current model of the environment used for the prediction
            density_estimate: The current estimate of the data density without
                the current step k
            propsed_actions: An initial proposition of actions to take

        Returns:
            action: The chosen action
            next_proposed_actions: An initial proposition for future actions
            density_estimate: The updated density estimate now incorporating
                the current step k
        """

        if self.reuse_proposed_actions:
            n_random_starts = self.n_starts - 1

        expl_key, new_proposed_actions_key, expl_action_key, _ = jax.random.split(expl_key, 4)

        if n_random_starts > 0:
            random_proposed_actions = jax.random.uniform(
                key=new_proposed_actions_key, shape=(n_random_starts, *proposed_actions.shape), minval=-1, maxval=1
            )

            if self.reuse_proposed_actions:
                all_proposed_actions = jnp.concatenate([proposed_actions[None, :], random_proposed_actions], axis=0)
            else:
                all_proposed_actions = random_proposed_actions
        else:
            all_proposed_actions = proposed_actions[None, :]

        proposed_actions, loss = optimize_actions_multistart(
            loss_function=self.loss_function,
            grad_loss_function=self.grad_loss_function,
            all_proposed_actions=all_proposed_actions,
            model=model,
            optimizer=self.excitation_optimizer,
            init_obs=obs,
            density_estimate=density_estimate,
            n_opt_steps=self.n_opt_steps,
            tau=self.tau,
            target_distribution=self.target_distribution,
            rho_obs=self.rho_obs,
            rho_act=self.rho_act,
            penalty_order=self.penalty_order,
        )

        action = proposed_actions[0, :]
        if self.clip_action:
            action = jnp.clip(action, -1, 1)

        next_proposed_actions = proposed_actions.at[:-1, :].set(proposed_actions[1:, :])

        new_proposed_action = jax.random.uniform(key=expl_action_key, minval=-1, maxval=1)
        next_proposed_actions = next_proposed_actions.at[-1, :].set(new_proposed_action)

        # update grid KDE with x_k and u_k
        density_estimate = update_density_estimate_single_observation(
            density_estimate, jnp.concatenate([obs, action], axis=-1)
        )

        return action, next_proposed_actions, density_estimate, loss, expl_key



from typing import Tuple
import matplotlib.pyplot as plt
from tqdm import tqdm
import numpy as np

import jax
import jax.numpy as jnp
import equinox as eqx
import optax

import exciting_environments as excenvs
from dmpe.algorithms.algorithm_utils import interact_and_observe, default_dmpe_parameterization
from dmpe.evaluation.plotting_utils import plot_sequence_and_prediction
from dmpe.excitation import loss_function, Exciter
from dmpe.models.model_training import ModelTrainer
from dmpe.utils.density_estimation import DensityEstimate, build_grid
from dmpe.utils.metrics import JSDLoss


def excite_and_fit(
    n_time_steps: int,
    env: excenvs.CoreEnvironment,
    model: eqx.Module,
    obs: jax.Array,
    state: excenvs.ClassicCoreEnvironment.State,
    proposed_actions: jax.Array,
    exciter: Exciter,
    model_trainer: ModelTrainer,
    density_estimate: DensityEstimate,
    observations: jax.Array,
    actions: jax.Array,
    opt_state_model: optax.OptState,
    loader_key: jax.random.PRNGKey,
    expl_key: jax.random.PRNGKey,
    plot_every: int,
) -> Tuple[jax.Array, jax.Array, eqx.Module, DensityEstimate]:
    """
    Main algorithm to apply to a given (unknown) system and generate informative data from that system.

    Args:
        n_time_steps (int): The number of time steps to run the algorithm for.
        env (excenvs.CoreEnvironment): The environment object representing the system.
        model (eqx.Module): The model used for prediction.
        obs (jax.Array): The initial observation of the system.
        state (excenvs.ClassicCoreEnvironment.State): The initial state of the system.
        proposed_actions (jax.Array): The proposed actions for exploration.
        exciter (Exciter): The exciter object responsible for choosing actions.
        model_trainer (ModelTrainer): The model trainer object responsible for training the model.
        density_estimate (DensityEstimate): The density estimate used for exploration.
        observations (jax.Array): The history of observations.
        actions (jax.Array): The history of actions.
        opt_state_model (optax.OptState): The optimizer state for the model.
        loader_key (jax.random.PRNGKey): The key used for loading data.
        plot_every (int): The frequency at which to plot the sequence and prediction.

    Returns:
        Tuple[jax.Array, jax.Array, eqx.Module, DensityEstimate]: A tuple containing the history of observations,
        the history of actions, the updated model, and the updated density estimate.
    """
    prediction_losses = []
    data_losses = []

    for k in tqdm(range(n_time_steps)):
        action, proposed_actions, density_estimate, prediction_loss, expl_key = exciter.choose_action(
            obs=obs,
            model=model,
            density_estimate=density_estimate,
            proposed_actions=proposed_actions,
            expl_key=expl_key,
        )
        prediction_losses.append(prediction_loss)

        obs, state, actions, observations = interact_and_observe(
            env=env, k=jnp.array([k]), action=action, state=state, actions=actions, observations=observations
        )

        if model_trainer is not None:
            if k > model_trainer.start_learning:
                model, opt_state_model, loader_key = model_trainer.fit(
                    model=model,
                    k=jnp.array([k]),
                    observations=observations,
                    actions=actions,
                    opt_state=opt_state_model,
                    loader_key=loader_key,
                )

        data_loss = JSDLoss(
            density_estimate.p / jnp.sum(density_estimate.p),
            exciter.target_distribution / jnp.sum(exciter.target_distribution),
        )
        data_losses.append(data_loss)

        if k % plot_every == 0 and k > 0:
            print("last input opt loss:", prediction_losses[-1])
            print("current data loss:", data_loss)
            fig, axs = plot_sequence_and_prediction(
                observations=observations[: k + 2, :],
                actions=actions[: k + 1, :],
                tau=exciter.tau,
                obs_labels=env.obs_description,
                actions_labels=[r"$u$"],
                model=model,
                init_obs=obs,
                proposed_actions=proposed_actions,
            )
            plt.show()

            plt.plot(np.log(data_losses))
            plt.grid(True)
            plt.show()

    return observations, actions, model, density_estimate, prediction_losses, proposed_actions


def excite_with_dmpe(
    env: excenvs.CoreEnvironment,
    exp_params: dict,
    proposed_actions: jax.Array,
    loader_key: jax.random.PRNGKey,
    expl_key: jax.random.PRNGKey,
    plot_every: bool | None = None,
):
    """
    Excite the system using the Differentiable Model Predictive Excitation (DMPE) algorithm.

    Args:
        env: The environment object representing the system.
        exp_params: The experiment parameters.
        proposed_actions: The proposed actions for exploration.
        model_key: The key for initializing the model.
        loader_key: The key used for loading data.
        expl_key: The key used for random action generation.
        plot_every: The frequency at which to plot the current data sequences.

    Returns:
        Tuple[jnp.ndarray, jnp.ndarray, eqx.Module, DensityEstimate]: A tuple containing the history of observations,
        the history of actions, the trained model, and the density estimate.
    """
    dim_obs_space = env.physical_state_dim  # assumes fully observable system
    dim_action_space = env.action_dim
    dim = dim_obs_space + dim_action_space
    n_grid_points = exp_params["alg_params"]["points_per_dim"] ** dim

    # setup x_0 / y_0
    obs, state = env.reset()
    obs = obs[0]

    # setup memory variables
    observations = jnp.zeros((exp_params["n_time_steps"], dim_obs_space))
    observations = observations.at[0].set(obs)
    actions = jnp.zeros((exp_params["n_time_steps"] - 1, dim_action_space))

    exciter = Exciter(
        loss_function=loss_function,
        grad_loss_function=jax.value_and_grad(loss_function, argnums=(2)),
        excitation_optimizer=optax.adabelief(exp_params["alg_params"]["action_lr"]),
        tau=env.tau,
        n_opt_steps=exp_params["alg_params"]["n_opt_steps"],
        target_distribution=jnp.ones(shape=(n_grid_points, 1)) * 1 / (1 - (-1)) ** dim,
        rho_obs=exp_params["alg_params"]["rho_obs"],
        rho_act=exp_params["alg_params"]["rho_act"],
        penalty_order=exp_params["alg_params"]["penalty_order"],
        clip_action=exp_params["alg_params"]["clip_action"],
        n_starts=exp_params["alg_params"]["n_starts"],
        reuse_proposed_actions=exp_params["alg_params"]["reuse_proposed_actions"],
    )

    if exp_params["model_trainer_params"] is None or exp_params["model_params"] is None:
        model_trainer = None
        model = exp_params["model_env_wrapper"](env)
        opt_state_model = None
    else:
        model_trainer = ModelTrainer(
            start_learning=exp_params["model_trainer_params"]["start_learning"],
            training_batch_size=exp_params["model_trainer_params"]["training_batch_size"],
            n_train_steps=exp_params["model_trainer_params"]["n_train_steps"],
            sequence_length=exp_params["model_trainer_params"]["sequence_length"],
            featurize=exp_params["model_trainer_params"]["featurize"],
            model_optimizer=optax.adabelief(exp_params["model_trainer_params"]["model_lr"]),
            tau=env.tau,
        )
        model = exp_params["model_class"](**exp_params["model_params"])
        opt_state_model = model_trainer.model_optimizer.init(eqx.filter(model, eqx.is_inexact_array))

    density_estimate = DensityEstimate(
        p=jnp.zeros([n_grid_points, 1]),
        x_g=build_grid(dim, low=-1, high=1, points_per_dim=exp_params["alg_params"]["points_per_dim"]),
        bandwidth=jnp.array([exp_params["alg_params"]["bandwidth"]]),
        n_observations=jnp.array([0]),
    )

    observations, actions, model, density_estimate, losses, proposed_actions = excite_and_fit(
        n_time_steps=exp_params["n_time_steps"],
        env=env,
        model=model,
        obs=obs,
        state=state,
        proposed_actions=proposed_actions,
        exciter=exciter,
        model_trainer=model_trainer,
        density_estimate=density_estimate,
        observations=observations,
        actions=actions,
        opt_state_model=opt_state_model,
        loader_key=loader_key,
        expl_key=expl_key,
        plot_every=plot_every if plot_every is not None else exp_params["n_time_steps"] + 1,
    )

    return observations, actions, model, density_estimate, losses, proposed_actions


def default_dmpe(env, seed=0, featurize=None, model_class=None, plot_every=None):
    """Runs DMPE with default parameterization. The parameter choices might
    not be optimal for a given system.

    In future work, automatic tuning for the parameters will be added such that no
    manual tuning is required.

    Args:
        env: The environment object representing the system.

    Returns:
        Tuple[jnp.ndarray, jnp.ndarray, eqx.Module, DensityEstimate]: A tuple containing the history of observations,
        the history of actions, the trained model, and the density estimate.
    """

    return excite_with_dmpe(
        env,
        *default_dmpe_parameterization(env, seed, featurize, model_class),
        plot_every,
    )



from typing import Tuple

import jax
import jax.numpy as jnp
import equinox as eqx
from haiku import PRNGSequence

import exciting_environments as excenvs
from dmpe.utils.signals import aprbs
from dmpe.utils.density_estimation import select_bandwidth
from dmpe.models import NeuralEulerODE


@eqx.filter_jit
def interact_and_observe(
    env: excenvs.CoreEnvironment,
    k: int,
    action: jax.Array,
    state: excenvs.ClassicCoreEnvironment.State,
    actions: jax.Array,
    observations: jax.Array,
) -> Tuple[jax.Array, excenvs.ClassicCoreEnvironment.State, jax.Array, jax.Array]:
    """
    Interact with the environment and store the action and the resulting observation.

    Args:
        env: The environment object.
        k: The current time step.
        action: The action to be taken at time step k.
        state: The state of the environment at time step k.
        actions: The list of actions taken so far.
        observations: The list of observations observed so far.

    Returns:
        obs: The updated observation at time step k+1.
        state: The updated state of the environment at time step k+1.
        actions: The updated list of actions taken so far.
        observations: The updated list of observations observed so far.
    """

    # apply u_k and go to x_{k+1}

    obs, state = env.step(state, action, env.env_properties)

    actions = actions.at[k].set(action)  # store u_k
    observations = observations.at[k + 1].set(obs)  # store x_{k+1}

    return obs, state, actions, observations


def default_dmpe_parameterization(env: excenvs.CoreEnvironment, seed: int = 0, featurize=None, model_class=None):
    """Returns a default parameterization for the DMPE algorithm.

    This parameterization is intended as a starting point to apply to a given system.
    The parameters are not necessarily optimal for any given system but should give a
    reasonable first impression. Currently, featurization of the model state e.g. angles
    needs to be provided manually.

    In future work, automatic tuning for the parameters will be added such that no
    manual tuning is required.

    Args:
        env (excenvs.CoreEnvironment): The environment object representing the system.
        seed (int): The seed for the random number generator.

    Returns:
        Tuple[Dict, jax.Array, jax.random.PRNGKey, jax.random.PRNGKey]: A tuple containing the experiment parameters,
        the initial proposed actions, the key for loading data in model learning, and the key for random action generation.

    """
    alg_params = dict(
        bandwidth=None,
        n_prediction_steps=50,
        points_per_dim=50,
        action_lr=1e-1,
        n_opt_steps=10,
        rho_obs=1,
        rho_act=1,
        penalty_order=2,
        clip_action=True,
        n_starts=5,
        reuse_proposed_actions=True,
    )
    alg_params["bandwidth"] = float(
        select_bandwidth(
            delta_x=2,
            dim=env.physical_state_dim + env.action_dim,
            n_g=alg_params["points_per_dim"],
            percentage=0.3,
        )
    )

    model_trainer_params = dict(
        start_learning=alg_params["n_prediction_steps"],
        training_batch_size=128,
        n_train_steps=3,
        sequence_length=alg_params["n_prediction_steps"],
        featurize=(lambda x: x) if featurize is None else featurize,
        model_lr=1e-4,
    )
    model_params = dict(obs_dim=env.physical_state_dim, action_dim=env.action_dim, width_size=128, depth=3, key=None)

    exp_params = dict(
        seed=seed,
        n_time_steps=15_000,
        model_class=NeuralEulerODE if model_class is None else model_class,
        env_params=None,
        alg_params=alg_params,
        model_trainer_params=model_trainer_params,
        model_params=model_params,
    )

    key = jax.random.PRNGKey(seed=exp_params["seed"])
    data_key, model_key, loader_key, expl_key, key = jax.random.split(key, 5)

    data_rng = PRNGSequence(data_key)
    exp_params["model_params"]["key"] = model_key

    # initial guess
    proposed_actions = jnp.hstack(
        [
            aprbs(alg_params["n_prediction_steps"], env.batch_size, 1, 10, next(data_rng))[0]
            for _ in range(env.action_dim)
        ]
    )

    return exp_params, proposed_actions, loader_key, expl_key



from operator import itemgetter

import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist

from pymoo.core.repair import Repair
from pymoo.core.problem import ElementwiseProblem
from pymoo.core.variable import Integer

from mixed_GA import Permutation


class StartFromZeroRepair(Repair):

    def _do(self, problem, X, **kwargs):

        print(X)

        I = np.where(X == 0)[1]

        for k in range(len(X)):
            i = I[k]
            X[k] = np.concatenate([X[k, i:], X[k, :i]])

        return X


def create_cities(n_cities, grid_width=100.0, grid_height=None, seed=None):
    if seed is not None:
        np.random.seed(seed)
    grid_height = grid_height if grid_height is not None else grid_width
    cities = np.random.random((n_cities, 2)) * [grid_width, grid_height]
    return cities


def visualize(problem, x, fig=None, ax=None, show=True, label=True):
    with plt.style.context("ggplot"):

        if fig is None or ax is None:
            fig, ax = plt.subplots()

        # plot cities using scatter plot
        ax.scatter(problem.cities[:, 0], problem.cities[:, 1], s=250)
        if label:
            # annotate cities
            for i, c in enumerate(problem.cities):
                ax.annotate(str(i), xy=c, fontsize=10, ha="center", va="center", color="white")

        # plot the line on the path
        for i in range(len(x)):
            current = x[i]
            next_ = x[(i + 1) % len(x)]
            ax.plot(problem.cities[[current, next_], 0], problem.cities[[current, next_], 1], "r--")

        fig.suptitle("Route length: %.4f" % problem.get_route_length(x))

        if show:
            fig.show()


class MultitaskingTravellingSalespersonProblem(ElementwiseProblem):

    def __init__(self, cities, **kwargs):
        """
        A two-dimensional traveling salesperson problem (TSP)

        Parameters
        ----------
        cities : numpy.array
            The cities with 2-dimensional coordinates provided by a matrix where where city is represented by a row.

        """
        n_cities, _ = cities.shape

        self.cities = cities
        self.D = cdist(cities, cities)

        variables = {number: Permutation(bounds=(0, n_cities)) for number in range(n_cities)}
        variables["x0"] = Integer(bounds=(-20, 20))
        variables["x1"] = Integer(bounds=(-20, 20))

        super(MultitaskingTravellingSalespersonProblem, self).__init__(vars=variables, n_obj=1)

        self.permutation_keys = []
        self.non_permutation_keys = []
        for key, value in variables.items():
            if isinstance(value, Permutation):
                self.permutation_keys.append(key)
            else:
                self.non_permutation_keys.append(key)

        self.permutation_keys = tuple(self.permutation_keys)
        self.non_permutation_keys = tuple(self.non_permutation_keys)

    def _evaluate(self, x, out, *args, **kwargs):

        x_perm = np.array(itemgetter(*self.permutation_keys)(x))
        x_nonperm = np.array(itemgetter(*self.non_permutation_keys)(x))

        out["F"] = self.get_route_length(x_perm) + self.rosenbrock_function(x_nonperm)

    def get_route_length(self, x):
        n_cities = len(x)
        dist = 0
        for k in range(n_cities - 1):
            i, j = x[k], x[k + 1]
            dist += self.D[i, j]

        last, first = x[-1], x[0]
        dist += self.D[last, first]  # back to the initial city
        return dist

    def rosenbrock_function(self, x):
        return (1 - x[0]) ** 2 + 100 * (x[1] - x[0] ** 2) ** 2



import jax
import jax.nn as jnn
import jax.numpy as jnp

import equinox as eqx
import diffrax


class MLP(eqx.Module):
    mlp: eqx.nn.MLP

    def __init__(self, obs_dim, action_dim, width_size, depth, *, key, **kwargs):
        super().__init__(**kwargs)
        self.mlp = eqx.nn.MLP(
            in_size=(obs_dim + action_dim),
            out_size=obs_dim,
            width_size=width_size,
            depth=depth,
            activation=jnn.leaky_relu,
            key=key,
        )

    def __call__(self, obs, action):
        obs_action = jnp.hstack([obs, action])
        return self.mlp(obs_action)


class NeuralODE(eqx.Module):
    func: MLP
    _solver: diffrax.AbstractSolver

    def __init__(self, solver, obs_dim, action_dim, width_size, depth, *, key, **kwargs):
        super().__init__(**kwargs)
        self.func = MLP(obs_dim, action_dim, width_size, depth, key=key)
        self._solver = solver

    def __call__(self, init_obs, actions, tau):

        args = (actions, None)

        def action_helper(t, args):
            actions = args
            return actions[jnp.array(t / tau, int)]

        def vector_field(t, y, args):
            actions, _ = args

            action = action_helper(t, actions)
            dy_dt = self.func(y, action)
            return tuple(dy_dt)

        term = diffrax.ODETerm(vector_field)
        t0 = 0
        t1 = tau * actions.shape[0]

        y0 = tuple(init_obs)
        saveat = diffrax.SaveAt(ts=jnp.linspace(t0, t1, 1 + int(t1 / tau)))
        solution = diffrax.diffeqsolve(term, self._solver, t0, t1, dt0=tau, y0=y0, args=args, saveat=saveat)

        return jnp.transpose(jnp.array(solution.ys))


class NeuralODEPendulum(NeuralODE):
    def __call__(self, init_obs, actions, tau):
        observations = super().__call__(init_obs, actions, tau)
        return jnp.stack([(((observations[..., 0] + 1) % 2) - 1), observations[..., 1]], axis=-1)


class NeuralEulerODE(eqx.Module):
    func: MLP

    def __init__(self, obs_dim, action_dim, width_size, depth, *, key, **kwargs):
        super().__init__(**kwargs)
        self.func = MLP(obs_dim, action_dim, width_size, depth, key=key)

    def step(self, obs, action, tau):
        next_obs = obs + tau * self.func(obs, action)
        return next_obs

    def __call__(self, init_obs, actions, tau):

        def body_fun(carry, action):
            obs = carry
            obs = self.step(obs, action, tau)
            return obs, obs

        _, observations = jax.lax.scan(body_fun, init_obs, actions)
        observations = jnp.concatenate([init_obs[None, :], observations], axis=0)
        return observations


class NeuralEulerODEPendulum(NeuralEulerODE):
    """Pendulum specific model that deals with the periodic properties of the angle information."""

    def step(self, obs, action, tau):
        next_obs = super().step(obs, action, tau)
        next_obs = jnp.stack([(((next_obs[..., 0] + 1) % 2) - 1), next_obs[..., 1]], axis=-1)
        return next_obs


class NeuralEulerODECartpole(NeuralEulerODE):
    """Cartpole specific model that deals with the periodic properties of the angle information."""

    def step(self, obs, action, tau):
        next_obs = super().step(obs, action, tau)
        next_obs = jnp.stack(
            [next_obs[..., 0], next_obs[..., 1], (((next_obs[..., 2] + 1) % 2) - 1), next_obs[..., 3]], axis=-1
        )
        return next_obs



from typing import Callable

import jax
import jax.numpy as jnp
import equinox as eqx
import optax

from dmpe.models.model_utils import simulate_ahead


@eqx.filter_grad
def grad_loss(model, true_obs, actions, tau, featurize):

    pred_obs = jax.vmap(simulate_ahead, in_axes=(None, 0, 0, None))(model, true_obs[:, 0, :], actions, tau)

    feat_pred_obs = jax.vmap(featurize, in_axes=(0))(pred_obs)
    feat_true_obs = jax.vmap(featurize, in_axes=(0))(true_obs)

    return jnp.mean((feat_pred_obs - feat_true_obs) ** 2)


@eqx.filter_jit
def make_step(model, observations, actions, tau, opt_state, featurize, optim):
    grads = grad_loss(model, observations, actions, tau, featurize)
    updates, opt_state = optim.update(grads, opt_state)
    model = eqx.apply_updates(model, updates)
    return model, opt_state


def dataloader(memory, batch_size, sequence_length, key):
    observations, actions = memory.values()
    observations = jnp.stack(observations, axis=0)
    actions = jnp.stack(actions, axis=0)
    dataset_size = observations.shape[0]

    assert actions.shape[0] == dataset_size - 1

    indices = jnp.arange(dataset_size - sequence_length)

    while True:
        starting_points = jax.random.choice(key=key, a=indices, shape=(batch_size,), replace=True)
        (key,) = jax.random.split(key, 1)

        slice = jnp.linspace(
            start=starting_points, stop=starting_points + sequence_length, num=sequence_length, dtype=int
        ).T

        batched_observations = observations[slice]
        batched_actions = actions[slice]

        yield tuple([batched_observations, batched_actions])


@eqx.filter_jit
def load_single_batch(observations_array, actions_array, starting_points, sequence_length):

    slice = jnp.linspace(
        start=starting_points, stop=starting_points + sequence_length, num=sequence_length, dtype=int
    ).T

    batched_observations = observations_array[slice]
    batched_actions = actions_array[slice]

    batched_observations = batched_observations[:, :, :]
    batched_actions = batched_actions[:, :-1, :]
    return batched_observations, batched_actions


@eqx.filter_jit
def precompute_starting_points(n_train_steps, k, sequence_length, training_batch_size, loader_key):
    index_normalized = jax.random.uniform(loader_key, shape=(n_train_steps, training_batch_size)) * (
        k + 1 - sequence_length
    )
    starting_points = index_normalized.astype(jnp.int32)
    (loader_key,) = jax.random.split(loader_key, 1)

    return starting_points, loader_key


@eqx.filter_jit
def fit(
    model, n_train_steps, starting_points, sequence_length, observations, actions, tau, featurize, optim, init_opt_state
):
    """Fit the model on the gathered data."""

    dynamic_init_model_state, static_model_state = eqx.partition(model, eqx.is_array)
    init_carry = (dynamic_init_model_state, init_opt_state)

    def body_fun(i, carry):
        dynamic_model_state, opt_state = carry
        model_state = eqx.combine(static_model_state, dynamic_model_state)

        batched_observations, batched_actions = load_single_batch(
            observations, actions, starting_points[i, ...], sequence_length
        )
        new_model_state, new_opt_state = make_step(
            model_state, batched_observations, batched_actions, tau, opt_state, featurize, optim
        )

        new_dynamic_model_state, new_static_model_state = eqx.partition(new_model_state, eqx.is_array)
        assert eqx.tree_equal(static_model_state, new_static_model_state) is True
        return (new_dynamic_model_state, new_opt_state)

    final_dynamic_model_state, final_opt_state = jax.lax.fori_loop(
        lower=0, upper=n_train_steps, body_fun=body_fun, init_val=init_carry
    )
    final_model = eqx.combine(static_model_state, final_dynamic_model_state)
    return final_model, final_opt_state


class ModelTrainer(eqx.Module):
    start_learning: jnp.int32
    training_batch_size: jnp.int32
    n_train_steps: jnp.int32
    sequence_length: jnp.int32
    featurize: Callable
    model_optimizer: optax._src.base.GradientTransformationExtraArgs
    tau: jnp.float32

    @eqx.filter_jit
    def fit(self, model, k, observations, actions, opt_state, loader_key):
        starting_points, loader_key = precompute_starting_points(
            n_train_steps=self.n_train_steps,
            k=k,
            sequence_length=self.sequence_length,
            training_batch_size=self.training_batch_size,
            loader_key=loader_key,
        )

        final_model, final_opt_state = fit(
            model=model,
            n_train_steps=self.n_train_steps,
            starting_points=starting_points,
            sequence_length=self.sequence_length,
            observations=observations,
            actions=actions,
            tau=self.tau,
            featurize=self.featurize,
            optim=self.model_optimizer,
            init_opt_state=opt_state,
        )
        return final_model, final_opt_state, loader_key



import jax
import jax.numpy as jnp

import equinox as eqx


def select_bandwidth(
    delta_x: float,
    dim: int,
    n_g: int,
    percentage: float,
):
    """Select a bandwidth for the kernel density estimate by a rough heuristic.

    The bandwidth is designed so that the kernel is still at a given percentage of
    its maximum value when a step is taken in each dimension of the underlying
    grid.

    Args:
        delta_x: The size of the space in each dimension.
        dim: The dimension of the space.
        n_g: Number of grid points per dimension.
        percentage: The percentage of the maximum value of the kernel at the other
            grid point reached by stepping once in each dimension on the grid.
    """
    return delta_x * jnp.sqrt(dim) / (n_g * jnp.sqrt(-2 * jnp.log(percentage)))


@jax.jit
def gaussian_kernel(x: jax.Array, bandwidth: float) -> jax.Array:
    """Evaluates the Gaussian RBF kernel at x with given bandwidth. This can take arbitrary
    dimensions for 'x' and will compute the output by broadcasting. The last dimension of
    the input needs to be the dimension of the data which is reduced along.
    """
    data_dim = x.shape[-1]
    factor = bandwidth**data_dim * jnp.power(2 * jnp.pi, data_dim / 2)
    return 1 / factor * jnp.exp(-jnp.linalg.norm(x, axis=-1) ** 2 / (2 * bandwidth**2))


class DensityEstimate(eqx.Module):
    """Holds an estimation of the density of sampled data points.

    Args:
        p: The probability estimates at the grid points
        x_g: The grid points
        bandwidth: The bandwidth of the kernel density estimate
        n_observations: The number of observations that make up the current
            estimate
    """

    p: jax.Array
    x_g: jax.Array
    bandwidth: jax.Array
    n_observations: jax.Array

    @classmethod
    def from_estimate(cls, p, n_additional_observations, density_estimate):
        """Create a density estimate recursively from an existing estimate."""

        return cls(
            p=p,
            n_observations=(density_estimate.n_observations + n_additional_observations),
            x_g=density_estimate.x_g,
            bandwidth=density_estimate.bandwidth,
        )

    @classmethod
    def from_dataset(
        cls, observations, actions, use_actions=True, points_per_dim=30, x_min=-1, x_max=1, bandwidth=0.05
    ):
        """Create a fresh density estimate from gathered data."""

        if use_actions:
            dim = observations.shape[-1] + actions.shape[-1]
        else:
            dim = observations.shape[-1]
        n_grid_points = points_per_dim**dim

        density_estimate = cls(
            p=jnp.zeros([1, n_grid_points, 1]),
            x_g=build_grid(dim, x_min, x_max, points_per_dim),
            bandwidth=jnp.array([bandwidth]),
            n_observations=jnp.array([0]),
        )

        if observations.shape[0] == actions.shape[0] + 1:
            data_points = (
                jnp.concatenate([observations[0:-1, :], actions], axis=-1)[None] if use_actions else observations[None]
            )
        else:
            data_points = jnp.concatenate([observations, actions], axis=-1)[None] if use_actions else observations[None]

        density_estimate = jax.vmap(
            update_density_estimate_multiple_observations,
            in_axes=(DensityEstimate(0, None, None, None), 0),
            out_axes=(DensityEstimate(0, None, None, None)),
        )(
            density_estimate,
            data_points,
        )
        return density_estimate


@jax.jit
def update_density_estimate_single_observation(
    density_estimate: DensityEstimate,
    data_point: jax.Array,
) -> jax.Array:
    """Recursive update to the kernel density estimation (KDE) on a fixed grid.

    Args:
        density_estimate: The density estimate before the update
        data_point: The new data point

    Returns:
        The updated density estimate
    """
    kernel_value = gaussian_kernel(x=density_estimate.x_g - data_point, bandwidth=density_estimate.bandwidth)
    p_est = (
        1
        / (density_estimate.n_observations + 1)
        * (density_estimate.n_observations * density_estimate.p + kernel_value[..., None])
    )

    return DensityEstimate.from_estimate(p=p_est, n_additional_observations=1, density_estimate=density_estimate)


@jax.jit
def update_density_estimate_multiple_observations(
    density_estimate: DensityEstimate,
    data_points: jax.Array,
) -> jax.Array:
    """Add a new sequence of data points to the current data density estimate.

    Args:
        density_estimate: The density estimate before the update
        data_points: The sequence of data_points

    Returns:
        The updated values for the density estimate
    """

    def shifted_gaussian_kernel(x, observation, bandwidth):
        return gaussian_kernel(x - observation, bandwidth)

    new_sum_part = jax.vmap(shifted_gaussian_kernel, in_axes=(None, 0, None))(
        density_estimate.x_g, data_points, density_estimate.bandwidth
    )
    new_sum_part = jnp.sum(new_sum_part, axis=0)[..., None]
    p_est = (
        1
        / (density_estimate.n_observations + data_points.shape[0])
        * (density_estimate.n_observations * density_estimate.p + new_sum_part)
    )

    return DensityEstimate.from_estimate(
        p=p_est, n_additional_observations=data_points.shape[0], density_estimate=density_estimate
    )


def build_grid(dim, low, high, points_per_dim):
    """Build a uniform grid of points in the given dimension."""
    xs = [jnp.linspace(low, high, points_per_dim) for _ in range(dim)]

    x_g = jnp.meshgrid(*xs)
    x_g = jnp.stack([_x for _x in x_g], axis=-1)
    x_g = x_g.reshape(-1, dim)

    assert x_g.shape[0] == points_per_dim**dim
    return x_g


def build_grid_2d(low, high, points_per_dim):
    return build_grid(2, low, high, points_per_dim)


def build_grid_3d(low, high, points_per_dim):
    return build_grid(3, low, high, points_per_dim)



from typing import Callable
import json
import pathlib
import glob

import matplotlib.pyplot as plt
import numpy as np
import jax.numpy as jnp

from dmpe.models.model_utils import load_model
from dmpe.evaluation.plotting_utils import plot_sequence, plot_model_performance
from dmpe.evaluation.metrics_utils import default_jsd, default_ae, default_mcudsa, default_ksfc


def get_experiment_ids(results_path: pathlib.Path):
    json_file_paths = glob.glob(str(results_path / pathlib.Path("*.json")))
    identifiers = set([pathlib.Path(path).stem.split("_", maxsplit=1)[-1] for path in json_file_paths])
    return sorted(list(identifiers))


def load_experiment_results(exp_id: str, results_path: pathlib.Path, model_class=None, to_array=True):
    with open(results_path / pathlib.Path(f"params_{exp_id}.json"), "rb") as fp:
        params = json.load(fp)

    with open(results_path / pathlib.Path(f"data_{exp_id}.json"), "rb") as fp:
        data = json.load(fp)

        if to_array:
            try:
                observations = jnp.array(data["observations"])
                actions = jnp.array(data["actions"])
            except:
                try:
                    observations = jnp.array(np.concatenate(data["observations"]))
                    actions = jnp.array(np.concatenate(data["actions"]))
                except:
                    obs_dim = np.array(data["observations"][0]).shape[0]

                    observations = np.stack(data["observations"][:-obs_dim], axis=0)
                    observations = jnp.array(
                        np.concatenate([observations, np.array(data["observations"][-obs_dim:])[None, ...]])
                    )
                    actions = jnp.array(np.stack(data["actions"]))
        else:
            observations = data["observations"]
            actions = data["actions"]

    if model_class is not None:
        model = load_model(results_path / pathlib.Path(f"model_{exp_id}.json"), model_class)
        return params, observations, actions, model
    else:
        return params, observations, actions, None


def evaluate_experiment_metrics(observations, actions, metrics, featurize=None):
    results = {}

    if metrics is None:
        metrics = {
            "jsd": default_jsd,
            "ae": default_ae,
            "mcudsa": default_mcudsa,
            "ksfc": default_ksfc,
        }

    for name, metric in metrics.items():
        results[name] = metric(observations, actions).item()

    if featurize is not None:
        assert isinstance(featurize, Callable)
        for name, metric in metrics.items():
            results[f"{name}_feat"] = metric(featurize(observations), actions).item()

    return results


def evaluate_algorithm_metrics(identifiers, results_path, featurize=None):
    results = {}
    for identifier in identifiers:
        _, observations, actions, _ = load_experiment_results(
            exp_id=identifier, results_path=results_path, model_class=None
        )
        single_result = evaluate_experiment_metrics(observations, actions, featurize=featurize)

        if len(results.keys()) == 0:
            for key, value in single_result.items():
                results[key] = [value]
        else:
            for key, value in single_result.items():
                results[key].append(value)
    return results


def evaluate_metrics(algorithm_names, n_results, results_parent_path, featurize):
    """Gathers the last 'n_results' experiments for differnet algorithms and evaluates the metrics."""

    results = {}
    for algorithm_name in algorithm_names:
        results_path = results_parent_path / pathlib.Path(algorithm_name)
        algorithm_results = evaluate_algorithm_metrics(
            identifiers=get_experiment_ids(results_path)[-n_results:],
            results_path=results_path,
            featurize=featurize,
        )
        results[algorithm_name] = algorithm_results
    return results


def extract_metrics_over_timesteps(experiment_ids, results_path, lengths, metrics=None, slotted=False):
    all_results = []
    for idx, identifier in enumerate(experiment_ids):
        print(f"Experiment {identifier} at index {idx}")

        _, observations, actions, _ = load_experiment_results(
            exp_id=identifier,
            results_path=results_path,
            model_class=None,
        )
        if slotted:
            single_results = []

            for i in range(len(lengths) - 1):
                single_results.append(
                    evaluate_experiment_metrics(
                        observations[lengths[i] : lengths[i + 1]], actions[lengths[i] : lengths[i + 1]], metrics=metrics
                    )
                )

        else:
            single_results = [
                evaluate_experiment_metrics(observations[:N], actions[:N], metrics=metrics) for N in lengths
            ]
        metric_keys = single_results[0].keys()

        results_by_metric = {key: [] for key in metric_keys}
        for result in single_results:
            for metric_key in metric_keys:
                results_by_metric[metric_key].append(result[metric_key])

        all_results.append(results_by_metric)

    print("Reshape to results by metric...")
    all_results_by_metric = {key: [] for key in metric_keys}
    for result in all_results:
        for metric_key in metric_keys:
            all_results_by_metric[metric_key].append(result[metric_key])

    for metric_key in all_results_by_metric.keys():
        all_results_by_metric[metric_key] = jnp.stack(jnp.array(all_results_by_metric[metric_key]))

    print("Done")

    return all_results_by_metric


def extract_metrics_over_timesteps_via_interpolation(experiment_ids, results_path, target_lengths, metrics=None):
    all_raw_results = []
    all_raw_lengths = []

    all_results = []
    for idx, identifier in enumerate(experiment_ids):
        print(f"Experiment {identifier} at index {idx} via interpolation")

        _, observations, actions, _ = load_experiment_results(
            exp_id=identifier,
            results_path=results_path,
            model_class=None,
            to_array=False,
        )

        raw_lengths = [len(subsequence) for subsequence in observations]
        raw_lengths = np.cumsum(raw_lengths[:-1])

        observations = jnp.array(np.concatenate(observations))
        actions = jnp.array(np.concatenate(actions))
        raw_results = [evaluate_experiment_metrics(observations[:N], actions[:N], metrics=metrics) for N in raw_lengths]

        metric_keys = raw_results[0].keys()

        raw_results_by_metric = {key: [] for key in metric_keys}
        for result in raw_results:
            for metric_key in metric_keys:
                raw_results_by_metric[metric_key].append(result[metric_key])

        all_raw_results.append(raw_results_by_metric)
        all_raw_lengths.append(raw_lengths)

        interpolated_results_by_metric = {}
        for metric_key in metric_keys:
            interpolated_results_by_metric[metric_key] = jnp.interp(
                x=target_lengths,
                xp=raw_lengths,
                fp=jnp.array(raw_results_by_metric[metric_key]),
            )
        all_results.append(interpolated_results_by_metric)

    print("Reshape to results by metric...")
    all_results_by_metric = {key: [] for key in metric_keys}
    for result in all_results:
        for metric_key in metric_keys:
            all_results_by_metric[metric_key].append(result[metric_key])

    for metric_key in all_results_by_metric.keys():
        all_results_by_metric[metric_key] = jnp.stack(jnp.array(all_results_by_metric[metric_key]))

    print("Done")

    return {"interp": all_results_by_metric, "raw": all_raw_results, "raw_lengths": all_raw_lengths}


def quick_eval(env, identifier, results_path, model_class=None):
    params, observations, actions, model = load_experiment_results(
        exp_id=identifier, results_path=results_path, model_class=model_class
    )

    print(identifier)
    print(params["alg_params"])

    if observations.shape[0] == actions.shape[0]:
        actions = actions[:-1]

    fig, axs = plot_sequence(
        observations=observations,
        actions=actions,
        tau=env.tau,
        obs_labels=env.obs_description,
        action_labels=[r"$u$"],
    )
    plt.show()

    if model is not None:
        fig, axs = plot_model_performance(
            model=model,
            true_observations=observations[:1000],
            actions=actions[:999],
            tau=env.tau,
            obs_labels=env.obs_description,
            action_labels=[r"$u$"],
        )
        plt.show()



from functools import partial
from tqdm import tqdm

import jax
import jax.numpy as jnp

from dmpe.utils.density_estimation import (
    DensityEstimate,
    build_grid,
    update_density_estimate_multiple_observations,
)
from dmpe.utils.metrics import (
    JSDLoss,
    audze_eglais,
    MC_uniform_sampling_distribution_approximation,
    kiss_space_filling_cost,
    blockwise_ksfc,
    blockwise_mcudsa,
)


def default_jsd(observations, actions, points_per_dim=50, bounds=(-1, 1), bandwidth=0.05):

    if observations.shape[0] == actions.shape[0] + 1:
        observations = observations[0:-1, :]

    data_points = jnp.concatenate([observations, actions], axis=-1)
    dim = data_points.shape[-1]
    n_grid_points = points_per_dim**dim

    density_estimate = DensityEstimate(
        p=jnp.zeros([n_grid_points, 1]),
        x_g=build_grid(dim, low=bounds[0], high=bounds[1], points_per_dim=points_per_dim),
        bandwidth=jnp.array([bandwidth]),
        n_observations=jnp.array([0]),
    )

    if data_points.shape[0] > 5000:
        # if there are too many datapoints at once, split them up and add
        # them in smaller chunks to the density estimate

        block_size = 1000

        for n in range(0, data_points.shape[0] + 1, block_size):
            density_estimate = update_density_estimate_multiple_observations(
                density_estimate,
                data_points[n : min(n + block_size, data_points.shape[0])],
            )
    else:
        density_estimate = update_density_estimate_multiple_observations(
            density_estimate,
            data_points,
        )

    target_distribution = jnp.ones(density_estimate.p.shape)
    target_distribution /= jnp.sum(target_distribution)

    return JSDLoss(
        p=density_estimate.p / jnp.sum(density_estimate.p),
        q=target_distribution,
    )


def default_ae(observations, actions):
    if observations.shape[0] == actions.shape[0] + 1:
        observations = observations[0:-1, :]

    return audze_eglais(jnp.concatenate([observations, actions], axis=-1))


def default_mcudsa(observations, actions, bounds=(-1, 1), points_per_dim=20):
    if observations.shape[0] == actions.shape[0] + 1:
        observations = observations[0:-1, :]

    data_points = jnp.concatenate([observations, actions], axis=-1)
    dim = data_points.shape[-1]

    support_points = build_grid(dim, low=bounds[0], high=bounds[1], points_per_dim=points_per_dim)

    if dim > 2:
        return blockwise_mcudsa(data_points=data_points, support_points=support_points)
    else:
        return MC_uniform_sampling_distribution_approximation(data_points=data_points, support_points=support_points)


def default_ksfc(observations, actions, points_per_dim=20, bounds=(-1, 1), variance=0.01, eps=1e-6):
    if observations.shape[0] == actions.shape[0] + 1:
        observations = observations[0:-1, :]

    data_points = jnp.concatenate([observations, actions], axis=-1)
    dim = data_points.shape[-1]

    support_points = build_grid(dim, low=bounds[0], high=bounds[1], points_per_dim=points_per_dim)

    if dim > 2:
        return blockwise_ksfc(
            data_points=data_points, support_points=support_points, variances=jnp.ones([dim]) * variance, eps=eps
        )
    else:
        return kiss_space_filling_cost(
            data_points=data_points, support_points=support_points, variances=jnp.ones([dim]) * variance, eps=eps
        )



from operator import itemgetter
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats.qmc import LatinHypercube
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize

import dmpe
from dmpe.related_work.np_reimpl.env_utils import simulate_ahead_with_env
from dmpe.related_work.np_reimpl.metrics import (
    MNNS_without_penalty,
    audze_eglais,
)
from dmpe.related_work.mixed_GA import Permutation, Integer, Real

import exciting_environments as excenvs


def latin_hypercube_sampling(d, n, rng):
    """Samples random points with latin hypercube sampling and normalizes between -1 and 1."""
    return LatinHypercube(d=d, seed=rng).random(n=n) * 2 - 1


def soft_penalty(a, a_max=1, penalty_order=2):
    """Computes penalty for the given input. Assumes symmetry in all dimensions."""
    relued_a = np.maximum(np.abs(a) - a_max, np.zeros(a.shape))

    relued_a = relued_a**penalty_order

    penalty = np.sum(relued_a, axis=(-2, -1))
    return np.squeeze(penalty)


def generate_aprbs(amplitudes, durations):
    """Parameterizable aprbs. This is used to transform the aprbs parameters into a signal."""
    return np.concatenate([np.ones(duration) * amplitude for (amplitude, duration) in zip(amplitudes, durations)])


def compress_datapoints(datapoints, N_c, feature_dimension, dist_th):
    """
    Compresses a sequence of datapoints based for the GOATS algorthims.

    Args:
        datapoints (ndarray): The sequence of datapoints to be compressed.
        N_c (int): The number of extra points per sequence.
        feature_dimension (int): The index of the feature dimension in the datapoints array.
        dist_th (float): The threshold distance for determining the number of extra points.

    Returns:
        compressed_data (ndarray): The compressed sequence of datapoints.
        indices (list): The indices of the selected datapoints in the original sequence.

    """
    # split data
    considered_data = datapoints[..., feature_dimension]
    support_mask = np.diff(np.sign(considered_data)) != 0
    support_mask = np.concatenate([support_mask, np.ones(1, dtype=bool)], axis=0)

    curv_approx = np.diff(considered_data)
    curv_support_mask = np.diff(np.sign(curv_approx)) != 0
    curv_support_mask = np.concatenate([curv_support_mask, np.ones(2, dtype=bool)], axis=0)
    support_mask = np.logical_or(curv_support_mask, support_mask)

    support = considered_data[support_mask]
    support_points = datapoints[support_mask]

    # compute number of extra points per subsequence
    support_distances = np.concatenate([np.diff(support), np.zeros(1)])
    support_abs_distances = np.abs(support_distances)
    n_per_subsequence = np.zeros(support_distances.shape)

    n_per_subsequence[support_abs_distances > dist_th] = support_abs_distances[support_abs_distances > dist_th]
    n_per_subsequence *= N_c / np.sum(n_per_subsequence)
    n_per_subsequence = n_per_subsequence.astype(np.int32)

    # bring data together
    support_indices = np.where(support_mask == 1)[0]

    compressed_data = []
    indices = []

    for idx, (start, start_point, n_new_points, distance) in enumerate(
        zip(support, support_points, n_per_subsequence, support_distances)
    ):
        compressed_data.append(start_point)
        indices.append(support_indices[idx])

        if idx + 1 >= support.shape[0]:
            available_support_indices = np.arange(support_indices[idx] + 1, support.shape[0])
        else:
            available_support_indices = np.arange(support_indices[idx] + 1, support_indices[idx + 1])
        available_support = considered_data[available_support_indices]

        if n_new_points > 0 and len(available_support) > 0:
            new_samples = np.linspace(start, distance + start, n_new_points + 2)[1:-1]
            for sample in new_samples:
                dist = np.abs(sample - available_support)
                chosen_idx = np.argmin(dist)
                chosen_obs = datapoints[available_support_indices[chosen_idx]]
                if chosen_idx not in indices:
                    compressed_data.append(chosen_obs)
                    indices.append(available_support_indices[chosen_idx])

    compressed_data = np.stack(compressed_data)

    return compressed_data, indices


class ContinuousGoatsProblem(ElementwiseProblem):
    """pymoo-API optimization problem for the iGOATs algorithm.

    Optimizes amplitude permutations and durations of each specific amplitude.
    The amplitude levels are chosen beforehand.
    """

    def __init__(
        self,
        prediction_horizon,
        env,
        obs,
        env_state,
        featurize,
        bounds_amplitude,
        bounds_duration,
        starting_observations,
        starting_actions,
        compress_data,
        compression_target_N,
        rho_obs,
        rho_act,
        penalty_order,
        compression_feat_dim,
        compression_dist_th,
    ):

        self.env = env
        self.obs = obs
        self.env_state = env_state
        self.featurize = featurize

        amplitude_variables = {f"a_{number}": Real(bounds=bounds_amplitude) for number in range(prediction_horizon)}
        duration_variables = {f"d_{number}": Integer(bounds=bounds_duration) for number in range(prediction_horizon)}
        all_vars = dict(amplitude_variables, **duration_variables)

        self.permutation_keys = tuple()
        self.non_permutation_keys = tuple(all_vars.keys())

        super().__init__(
            vars=all_vars,
            n_obj=1,
        )

        self.prediction_horizon = prediction_horizon

        if starting_observations is not None and starting_actions is not None:
            starting_observations = featurize(np.stack(starting_observations))
            starting_actions = np.stack(starting_actions)

            self.starting_feat_datapoints = np.concatenate([starting_observations, starting_actions], axis=-1)
            if compress_data and len(starting_observations) > 2000:
                self.starting_feat_datapoints, _ = compress_datapoints(
                    self.starting_feat_datapoints,
                    N_c=int(compression_target_N),
                    feature_dimension=compression_feat_dim,
                    dist_th=compression_dist_th,
                )

        else:
            self.starting_feat_datapoints = None

        self.compress_data = compress_data
        self.compression_target_N = compression_target_N
        self.rho_obs = rho_obs
        self.rho_act = rho_act
        self.penalty_order = penalty_order
        self.compression_feat_dim = compression_feat_dim
        self.compression_dist_th = compression_dist_th

    def _evaluate(self, x, out, *args, **kwargs):
        action_parameters = np.fromiter(x.values(), dtype=np.float64)
        actions = generate_aprbs(
            amplitudes=action_parameters[: self.prediction_horizon],
            durations=action_parameters[self.prediction_horizon :].astype(np.int32),
        )[:, None]

        observations, _ = simulate_ahead_with_env(
            self.env,
            self.obs,
            self.env_state,
            actions,
        )
        feat_observations = self.featurize(observations)
        new_datapoints = np.concatenate([feat_observations[:-1], actions], axis=-1)

        if self.starting_feat_datapoints is None:
            score = audze_eglais(new_datapoints)
        else:
            score = MNNS_without_penalty(
                data_points=self.starting_feat_datapoints,
                new_data_points=new_datapoints,
            )

        penalty_terms = self.rho_obs * soft_penalty(
            a=observations, a_max=1, penalty_order=self.penalty_order
        ) + self.rho_act * soft_penalty(a=actions, a_max=1, penalty_order=self.penalty_order)
        out["F"] = np.squeeze(score).item() + penalty_terms.item()


def optimize_continuous_aprbs(
    opt_algorithm,
    prediction_horizon: int,
    application_horizon: int,
    env,
    obs: np.ndarray,
    env_state: np.ndarray,
    bounds_amplitude: tuple,
    bounds_duration: tuple,
    n_generations: int,
    featurize: Callable,
    rng: np.random.Generator,
    starting_observations: np.ndarray,
    starting_actions: np.ndarray,
    compress_data: bool,
    compression_target_N: int,
    rho_obs: float,
    rho_act: float,
    penalty_order: int,
    compression_feat_dim: int,
    compression_dist_th: float,
):
    """Optimize an APRBS signal with continuous amplitude levels for system excitiation."""

    opt_problem = ContinuousGoatsProblem(
        prediction_horizon,
        env,
        obs,
        env_state,
        featurize,
        bounds_amplitude,
        bounds_duration,
        starting_observations=starting_observations,
        starting_actions=starting_actions,
        compress_data=compress_data,
        compression_target_N=compression_target_N,
        rho_obs=rho_obs,
        rho_act=rho_act,
        penalty_order=penalty_order,
        compression_feat_dim=compression_feat_dim,
        compression_dist_th=compression_dist_th,
    )

    res = minimize(
        problem=opt_problem,
        algorithm=opt_algorithm,
        termination=("n_gen", n_generations),
        seed=rng.integers(low=0, high=2**32 - 1, size=1).item(),
        save_history=False,
        verbose=False,
    )
    proposed_aprbs_params = np.fromiter(res.X.values(), dtype=np.float64)

    amplitudes = proposed_aprbs_params[:application_horizon]
    all_durations = proposed_aprbs_params[prediction_horizon:]
    durations = all_durations[:application_horizon].astype(np.int32)
    new_actions = generate_aprbs(amplitudes=amplitudes, durations=durations)[:, None]

    new_observations, env_state = simulate_ahead_with_env(
        env,
        obs,
        env_state,
        new_actions,
    )
    return new_observations, new_actions, env_state


class GoatsProblem(ElementwiseProblem):
    """pymoo-API optimization problem for the GOATs and sGOATs algorithms.

    Optimizes amplitude permutations and durations of each specific amplitude.
    The amplitude levels are chosen beforehand.

    TODO: arbitrary observation and input dimensions
    """

    def __init__(
        self,
        amplitudes,
        env,
        obs,
        env_state,
        featurize,
        bounds_duration=(1, 50),
        starting_observations=None,
        starting_actions=None,
        compress_data: bool = True,
        compression_target_N: int = 500,
        rho_obs: float = 1e3,
        rho_act: float = 1e3,
        penalty_order: int = 2,
        compression_feat_dim: int = 0,
        compression_dist_th: float = 0.1,
        share_of_current_sequence: float = 1,
    ):

        n_amplitudes = amplitudes.shape[0]

        self.env = env
        self.obs = obs
        self.env_state = env_state
        self.featurize = featurize

        amplitude_variables = {f"a_{number}": Permutation(bounds=(0, n_amplitudes)) for number in range(n_amplitudes)}
        duration_variables = {f"d_{number}": Integer(bounds=bounds_duration) for number in range(n_amplitudes)}

        self.permutation_keys = tuple(amplitude_variables.keys())
        self.non_permutation_keys = tuple(duration_variables.keys())

        all_vars = dict(amplitude_variables, **duration_variables)

        super().__init__(
            vars=all_vars,
            n_obj=1,
        )

        self.amplitudes = amplitudes
        self.n_amplitudes = n_amplitudes
        if starting_observations is not None and starting_actions is not None:
            starting_observations = featurize(starting_observations)
            starting_actions = starting_actions

            self.starting_feat_datapoints = np.concatenate([starting_observations, starting_actions], axis=-1)
            if compress_data and len(self.starting_feat_datapoints) > 2000:
                self.starting_feat_datapoints, _ = compress_datapoints(
                    self.starting_feat_datapoints,
                    N_c=int(compression_target_N * (1 - share_of_current_sequence)),
                    feature_dimension=compression_feat_dim,
                    dist_th=compression_dist_th,
                )
        else:
            self.starting_feat_datapoints = None

        self.compress_data = compress_data
        self.compression_target_N = compression_target_N
        self.rho_obs = rho_obs
        self.rho_act = rho_act
        self.penalty_order = penalty_order
        self.compression_feat_dim = compression_feat_dim
        self.compression_dist_th = compression_dist_th
        self.share_of_current_sequence = share_of_current_sequence

    def _evaluate(self, x, out, *args, **kwargs):
        indices = np.array(itemgetter(*self.permutation_keys)(x))
        durations = np.array(itemgetter(*self.non_permutation_keys)(x))

        applied_amplitudes = self.amplitudes[indices]

        actions = generate_aprbs(amplitudes=applied_amplitudes, durations=durations)[:, None]

        if isinstance(self.env, excenvs.core_env.CoreEnvironment):
            observations, _ = dmpe.models.model_utils.simulate_ahead_with_env(
                self.env,
                self.obs,
                self.env_state,
                actions,
            )
        else:
            observations, _ = simulate_ahead_with_env(
                self.env,
                self.obs,
                self.env_state,
                actions,
            )

        feat_observations = self.featurize(observations)
        feat_datapoints = np.concatenate([feat_observations[:-1], actions], axis=-1)

        if self.compress_data:
            feat_datapoints, indices = compress_datapoints(
                feat_datapoints,
                N_c=int(self.compression_target_N * self.share_of_current_sequence),
                feature_dimension=self.compression_feat_dim,
                dist_th=self.compression_dist_th,
            )

        if self.starting_feat_datapoints is not None:
            feat_datapoints = np.concatenate([self.starting_feat_datapoints, feat_datapoints])

        score = audze_eglais(feat_datapoints)

        # TODO: should the number of steps be included in the loss? This is to incite shorter trajectories..?
        N = feat_datapoints.shape[0]

        penalty_terms = self.rho_obs * soft_penalty(
            a=observations, a_max=1, penalty_order=self.penalty_order
        ) + self.rho_act * soft_penalty(a=actions, a_max=1, penalty_order=self.penalty_order)

        out["F"] = N * score + penalty_terms.item()


def optimize_permutation_aprbs(
    opt_algorithm,
    amplitudes: np.ndarray,
    env,
    obs: np.ndarray,
    env_state: np.ndarray,
    bounds_duration: tuple,
    n_generations: int,
    featurize: Callable,
    rng: np.random.Generator,
    verbose: bool,
    starting_observations: np.ndarray | None,
    starting_actions: np.ndarray | None,
    compress_data: bool,
    compression_target_N: int,
    rho_obs: float,
    rho_act: float,
    penalty_order: int,
    compression_feat_dim: int,
    compression_dist_th: float,
    share_of_current_sequence: float,
):
    """Optimize an APRBS signal with predefined amplitude levels for system excitiation."""

    opt_problem = GoatsProblem(
        amplitudes=amplitudes,
        env=env,
        obs=obs,
        env_state=env_state,
        featurize=featurize,
        bounds_duration=bounds_duration,
        starting_observations=starting_observations,
        starting_actions=starting_actions,
        compress_data=compress_data,
        compression_target_N=compression_target_N,
        rho_act=rho_act,
        rho_obs=rho_obs,
        penalty_order=penalty_order,
        compression_dist_th=compression_dist_th,
        compression_feat_dim=compression_feat_dim,
        share_of_current_sequence=share_of_current_sequence,
    )

    res = minimize(
        problem=opt_problem,
        algorithm=opt_algorithm,
        termination=("n_gen", n_generations),
        seed=rng.integers(low=0, high=2**32 - 1, size=1).item(),
        save_history=False,
        verbose=verbose,
    )

    indices = np.array(itemgetter(*opt_problem.permutation_keys)(res.X))
    applied_amplitudes = opt_problem.amplitudes[indices]

    applied_durations = np.array(itemgetter(*opt_problem.non_permutation_keys)(res.X))

    actions = generate_aprbs(amplitudes=applied_amplitudes, durations=applied_durations)[:, None]

    observations, last_env_state = simulate_ahead_with_env(
        env,
        obs,
        env_state,
        actions,
    )

    return observations, actions, last_env_state



"""Re-implementation of metrics using pure numpy instead of jax.

This is necessary for a pure numpy implementation of the related work (which cannot be efficiently implemented in jax).
"""

import numpy as np


def KLDLoss(p: np.ndarray, q: np.ndarray):
    """Computes the sample KLD between two inputs.

    The last dim of the input needs to be of length 1. The summation occurs along the second to
    last dimension. All dimensions before that are kept as they are. Overall the shape of the
    two inputs must be indentical.

    TODO: add an eps=1e-32 to remove zero issues?
    """
    assert p.shape == q.shape, "The two inputs need to be of the same shape."
    assert p.shape[-1] == q.shape[-1] == 1, "Last dim needs to be of length 1 for PDFs"

    eps = 1e-12

    kld = (p + eps) * np.log((p + eps) / (q + eps))
    return np.squeeze(np.sum(kld, axis=-2))


def JSDLoss(p: np.ndarray, q: np.ndarray):
    """Computes the sample JSD between two inputs.

    The last dim of the input needs to be of length 1. The summation occurs along the second to
    last dimension. All dimensions before that are kept as they are. Overall the shape of the
    two inputs must be indentical.
    """
    assert p.shape == q.shape, "The two inputs need to be of the same shape."
    assert p.shape[-1] == q.shape[-1] == 1, "Last dim needs to be of length 1 for PDFs"

    m = (p + q) / 2
    return np.squeeze((KLDLoss(p, m) + KLDLoss(q, m)) / 2)


def MNNS_without_penalty(data_points: np.ndarray, new_data_points: np.ndarray) -> np.ndarray:
    """From [Smits2024].

    Implementation inspired by https://github.com/google/jax/discussions/9813

    TODO: Not sure about this penalty. Seems difficult to use for continuous action-spaces?
    They used quantized amplitude levels in their implementation.
    """
    L = new_data_points.shape[0]
    distance_matrix = np.linalg.norm(data_points[:, None, :] - new_data_points[None, ...], axis=-1)
    minimal_distances = np.min(distance_matrix, axis=0)
    return -np.sum(minimal_distances) / L


def average_euclidean_distance(data_points: np.ndarray):
    """From [Smits2024].

    Used within the MNNS penalty term.
    """
    N = data_points.shape[0]
    distance_matrix = np.linalg.norm(data_points[:, None, :] - data_points[None, ...], axis=-1)
    distances = distance_matrix[np.triu_indices(N, k=1)]
    return 2 / (N * (N - 1)) * np.sum(distances**2)


def MNNS(data_points: np.ndarray, new_data_points: np.ndarray, k_d_max=1, delta=None):
    """From [Smits2024].

    Implementation inspired by https://github.com/google/jax/discussions/9813

    Maximum nearest neighbor sequence. If used during optimization, the delta should
    be precomputed for the data_points.
    """
    score_without_penalty = MNNS_without_penalty(data_points, new_data_points)

    if delta is None:
        delta = average_euclidean_distance(data_points)

    raise NotImplementedError
    # action_counts = None
    # penalty = action_counts * k_d_max * delta

    # return score_without_penalty + penalty


def MC_uniform_sampling_distribution_approximation(data_points: np.ndarray, support_points: np.ndarray) -> np.ndarray:
    """From [Smits2024]. The minimax-design tries to minimize
    the distances of the data points to the support points.

    What stops the data points to just flock to a single support point?
    This is just looking at the shortest distance.
    """
    M = support_points.shape[0]
    distance_matrix = np.linalg.norm(data_points[:, None, :] - support_points[None, ...], axis=-1)
    minimal_distances = np.min(distance_matrix, axis=0)
    return np.sum(minimal_distances) / M


def audze_eglais(data_points: np.ndarray, eps: float = 0.001) -> np.ndarray:
    """From [Smits2024]. The maximin-desing penalizes points that
    are too close in the point distribution.

    TODO: There has to be a more efficient way to do this.
    """
    N = data_points.shape[0]
    distance_matrix = np.linalg.norm(data_points[:, None, :] - data_points[None, ...], axis=-1)
    distances = distance_matrix[np.triu_indices(N, k=1)]

    return 2 / (N * (N - 1)) * np.sum(1 / (distances**2 + eps))


def blockwise_mcudsa(data_points: np.ndarray, support_points: np.ndarray) -> np.ndarray:
    M = support_points.shape[0]

    block_size = 1_000
    value = np.zeros(1)

    for m in range(0, M, block_size):
        end = min(m + block_size, M)  # next block or until the end
        value = value + (
            MC_uniform_sampling_distribution_approximation(
                data_points=data_points,
                support_points=support_points[m:end],
            )
            * (end - m)  # denormalizing mean inside loss computation
            / M
        )

    return value


def kiss_space_filling_cost(
    data_points: np.ndarray,
    support_points: np.ndarray,
    variances: np.ndarray,
    eps: float = 1e-16,
) -> np.ndarray:
    """From [Kiss2024]. Slightly modified to use the mean instead of the sum in the denominator.
    The goal is to have the same metric value for identical data distributions with different number
    of data points.
    """
    difference = data_points[None, ...] - support_points[:, None, :]
    exponent = -0.5 * np.sum(difference**2 * 1 / variances, axis=-1)

    denominator = eps + np.mean(np.exp(exponent), axis=-1)

    return np.mean(1 / denominator, axis=0)


def blockwise_ksfc(
    data_points: np.ndarray,
    support_points: np.ndarray,
    variances: np.ndarray,
    eps: float = 1e-16,
) -> np.ndarray:
    M = support_points.shape[0]

    block_size = 1_000
    value = np.zeros(1)

    for m in range(0, M, block_size):
        end = min(m + block_size, M)  # next block or until the end
        value = value + (
            kiss_space_filling_cost(
                data_points=data_points,
                support_points=support_points[m:end],
                variances=variances,
                eps=eps,
            )
            * (end - m)  # denormalizing mean inside loss computation
            / M
        )

    return value
